import { Buffer } from 'buffer';
import EventEmitter from 'events';
function addMessageListener(socket, onMessage) {
    const chunks = [];
    let bufferLength = 0;
    let state = null;
    let triggerLength = 0;
    socket.on('data', (chunk) => {
        bufferLength += chunk.length;
        chunks.push(chunk);
        while (bufferLength >= triggerLength) {
            const bufferedChunk = chunks.length === 1 ? chunks[0] : Buffer.concat(chunks);
            const triggerChunk = bufferedChunk.subarray(0, triggerLength);
            bufferLength -= triggerLength;
            chunks.splice(0, chunks.length, bufferedChunk.subarray(triggerLength));
            switch (state) {
                case 'length':
                    state = 'content';
                    triggerLength = parseInt(triggerChunk.toString(), 16);
                    break;
                case 'content':
                    onMessage(triggerChunk);
                // falls through
                default:
                    state = 'length';
                    triggerLength = 9;
            }
        }
    });
}
// eslint-disable-next-line @typescript-eslint/no-empty-function
const emptyFunction = () => { };
function deferred() {
    let resolve = emptyFunction;
    let reject = emptyFunction;
    const promise = new Promise((res, rej) => {
        resolve = res;
        reject = rej;
    });
    return { resolve, reject, promise };
}
export class QuickJSDebugConnection extends EventEmitter {
    constructor(socket) {
        super();
        this.socket = socket;
        this.requestTimeout = 10000;
        this.requestVersion = 1;
        this.requestSeq = 1;
        this.requestReactions = new Map();
        addMessageListener(socket, (message) => {
            this.handleMessage(message);
        });
        socket.on('end', () => {
            this.emit('end');
            const reactions = [...this.requestReactions.values()];
            this.requestReactions.clear();
            reactions.forEach(({ reject }) => {
                reject(new Error('Protocol is closed'));
            });
            this.requestReactions.clear();
        });
        socket.on('error', (err) => this.emit('error', err));
    }
    close() {
        this.socket.end();
    }
    sendMessage(message) {
        const buffer = Buffer.from(`${JSON.stringify(message)}\n`);
        const lf = Buffer.from('\n');
        const packet = Buffer.concat([Buffer.from(buffer.length.toString(16).padStart(8, '0')), lf, buffer]);
        this.socket.write(packet);
    }
    sendEnvelope(type, data) {
        this.sendMessage({
            version: this.requestVersion,
            type,
            ...data
        });
    }
    sendRequestRaw(command, args) {
        const { requestSeq } = this;
        this.requestSeq += 1;
        this.sendEnvelope('request', {
            request: {
                request_seq: requestSeq,
                command,
                args
            }
        });
        return requestSeq;
    }
    sendRequest(command, args) {
        const requestSeq = this.sendRequestRaw(command, args);
        const timeoutError = new Error(`Request timeout ${this.requestTimeout}ms exceed.`);
        const { promise, resolve, reject } = deferred();
        this.requestReactions.set(requestSeq, { resolve: resolve, reject });
        const timeout = setTimeout(() => {
            reject(timeoutError);
        }, this.requestTimeout);
        return promise.finally(() => {
            clearTimeout(timeout);
        });
    }
    handleMessage(message) {
        const json = JSON.parse(message.toString());
        if (json.type === 'event') {
            const event = json.event;
            this.emit(event.type, event);
        }
        else if (json.type === 'response') {
            const response = json;
            const reaction = this.requestReactions.get(response.request_seq);
            if (reaction) {
                this.requestReactions.delete(response.request_seq);
                if (response.error) {
                    reaction.reject(new Error(response.error));
                }
                else {
                    reaction.resolve(response.body);
                }
            }
        }
    }
}
//# sourceMappingURL=connection.js.map