/// <reference types="node" resolution-mode="require"/>
import EventEmitter from 'events';
import { DebugProtocol } from '@vscode/debugprotocol';
import { DebugConnection } from './connection.js';
export interface StackFrameInfo {
    id: number;
    name: string;
    filename: string;
    line: number;
    column: number;
}
export declare class QuickJSStackFrame {
    session: QuickJSDebugSession;
    id: number;
    name: string;
    fileName: string;
    lineNumber: number;
    constructor(session: QuickJSDebugSession, frameInfo: StackFrameInfo);
    evaluateExpression<R = unknown>(expression: string): Promise<QuickJSVariable<R>>;
    evaluateHandle<T, R>(f: (args: T) => R, args?: T): Promise<QuickJSVariable<R>>;
    evaluateHandleGlobal<T, R>(f: (args: T) => R, args?: T): Promise<QuickJSVariable<R>>;
    evaluate<T, R>(f: (args: T) => R, args?: T): Promise<R>;
    evaluateGlobal<T, R>(f: (args: T) => R, args?: T): Promise<R>;
    getScopes(): Promise<QuickJSScope[]>;
}
export interface InspectOptions {
    maxDepth?: number;
    inspectProto?: boolean;
}
export declare class QuickJSHandle<T = unknown> {
    session: QuickJSDebugSession;
    name: string;
    ref: number;
    primitive?: boolean;
    primitiveValue?: unknown;
    type?: string;
    isArray?: boolean;
    indexedCount?: number;
    valueAsString?: string;
    constructor(session: QuickJSDebugSession, reference: number);
    getProperties(options?: Omit<DebugProtocol.VariablesArguments, 'variablesReference'>): Promise<QuickJSVariable<unknown>[]>;
    inspect(options?: InspectOptions): Promise<T>;
    private inspectInternal;
    toString(): string;
    valueOf(): {} | undefined;
    equals(x: QuickJSHandle | null | undefined): boolean;
}
export interface ScopeInfo {
    name: string;
    reference: number;
    expensive: boolean;
}
export declare class QuickJSScope extends QuickJSHandle<Record<string, unknown>> {
    expensive: boolean;
    constructor(session: QuickJSDebugSession, scopeInfo: ScopeInfo);
    toString(): string;
}
export interface VariableInfo {
    name: string;
    value: string;
    type?: string;
    variablesReference: number;
    indexedVariables?: number;
}
export declare class QuickJSVariable<T = unknown> extends QuickJSHandle<T> {
    constructor(session: QuickJSDebugSession, variableInfo: VariableInfo);
}
export interface BreakpointInfo {
    line: number;
    column?: number;
}
export interface StoppedEvent {
    thread: number;
    reason: 'entry' | 'exception' | 'breakpoint' | 'pause' | 'step' | 'stepIn' | 'stepOut';
}
export interface ContextEvent {
    thread: number;
    reason: 'new' | 'exited';
}
export declare class QuickJSDebugSession extends EventEmitter {
    connection: DebugConnection;
    constructor(connection: DebugConnection);
    continue(): Promise<{
        allThreadsContinued?: boolean | undefined;
    }>;
    pause(): Promise<void>;
    stepNext(): Promise<void>;
    stepIn(): Promise<void>;
    stepOut(): Promise<void>;
    evaluate<R = unknown>(frameId: number, expression: string): Promise<QuickJSVariable<R>>;
    traceStack(): Promise<QuickJSStackFrame[]>;
    getTopStack(): Promise<QuickJSStackFrame>;
    getScopes(frameId: number): Promise<QuickJSScope[]>;
    inspectVariable<T = unknown>(reference: number, options?: Omit<DebugProtocol.VariablesArguments, 'variablesReference'>): Promise<QuickJSVariable<T>[]>;
    resume(): void;
    setBreakpoints(fileName: string, breakpoints: BreakpointInfo[]): void;
    setStopOnException(enabled: boolean): void;
}
export interface QuickJSDebugSessionEventMap {
    stopped: (event: StoppedEvent) => void;
    context: (event: ContextEvent) => void;
    end: () => void;
}
export interface QuickJSDebugSession {
    on(eventName: 'stopped', listener: (event: StoppedEvent) => void): this;
    on(eventName: 'context', listener: (event: ContextEvent) => void): this;
    on(eventName: 'end', listener: () => void): this;
    once(eventName: 'stopped', listener: (event: StoppedEvent) => void): this;
    once(eventName: 'context', listener: (event: ContextEvent) => void): this;
    once(eventName: 'end', listener: () => void): this;
    off(eventName: 'stopped', listener: (event: StoppedEvent) => void): this;
    off(eventName: 'context', listener: (event: ContextEvent) => void): this;
    off(eventName: 'end', listener: () => void): this;
    addListener(eventName: 'stopped', listener: (event: StoppedEvent) => void): this;
    addListener(eventName: 'context', listener: (event: ContextEvent) => void): this;
    addListener(eventName: 'end', listener: () => void): this;
    removeListener(eventName: 'stopped', listener: (event: StoppedEvent) => void): this;
    removeListener(eventName: 'context', listener: (event: ContextEvent) => void): this;
    removeListener(eventName: 'end', listener: () => void): this;
    emit(eventName: 'stopped', event: StoppedEvent): boolean;
    emit(eventName: 'context', event: ContextEvent): boolean;
    emit(eventName: 'end'): boolean;
}
