import EventEmitter from 'events';
function generateFunctionCode(f, args, type) {
    if (typeof f === 'function') {
        const serializedArgs = JSON.stringify(args);
        if (type === 'eval') {
            return `(${String(f)})(${serializedArgs})`;
        }
        else {
            const serializedCode = JSON.stringify(`return (${String(f)})(arguments[0])`);
            return `(new Function(${serializedCode}))(${serializedArgs})`;
        }
    }
    return String(f);
}
export class QuickJSStackFrame {
    constructor(session, frameInfo) {
        this.session = session;
        this.id = frameInfo.id;
        this.name = frameInfo.name;
        this.fileName = frameInfo.filename;
        this.lineNumber = frameInfo.line;
    }
    async evaluateExpression(expression) {
        return this.session.evaluate(this.id, expression);
    }
    async evaluateHandle(f, args) {
        return this.evaluateExpression(generateFunctionCode(f, args, 'eval'));
    }
    async evaluateHandleGlobal(f, args) {
        return this.evaluateExpression(generateFunctionCode(f, args, 'function'));
    }
    async evaluate(f, args) {
        return (await this.evaluateHandle(f, args)).inspect();
    }
    async evaluateGlobal(f, args) {
        return (await this.evaluateHandleGlobal(f, args)).inspect();
    }
    async getScopes() {
        return this.session.getScopes(this.id);
    }
}
const QuickJSRef = Symbol('QuickJSRef');
export class QuickJSHandle {
    constructor(session, reference) {
        this.session = session;
        this.ref = reference;
        this.name = `#${reference}`;
    }
    async getProperties(options) {
        return this.session.inspectVariable(this.ref, options);
    }
    async inspect(options) {
        const referenceMap = new Map();
        const { maxDepth, inspectProto } = options ?? {};
        return this.inspectInternal(maxDepth ?? 16, {
            referenceMap,
            inspectProto
        });
    }
    async inspectInternal(depth, options) {
        const { referenceMap, inspectProto } = options;
        if (this.primitive) {
            return this.primitiveValue;
        }
        if (referenceMap.has(this.ref)) {
            return referenceMap.get(this.ref);
        }
        if (this.type === 'object') {
            if (depth > 0) {
                let result;
                let getPropOptions;
                let properties;
                if (this.isArray) {
                    result = [];
                    getPropOptions = {
                        filter: 'indexed',
                        start: 0,
                        count: this.indexedCount
                    };
                }
                else {
                    result = {};
                }
                referenceMap.set(this.ref, result);
                try {
                    properties = await this.getProperties(getPropOptions);
                }
                catch (err) {
                    properties = [];
                }
                await Promise.all(properties.map(async (property) => {
                    if (property.name === '__proto__') {
                        if (inspectProto) {
                            const proto = await property.inspectInternal(depth - 1, options);
                            if (typeof proto === 'object') {
                                Object.setPrototypeOf(result, proto);
                            }
                        }
                    }
                    else {
                        result[property.name] = await property.inspectInternal(depth - 1, options);
                    }
                }));
                result[QuickJSRef] = this.ref;
                return result;
            }
        }
        return String(this);
    }
    toString() {
        return this.valueAsString ?? String(this.primitiveValue);
    }
    valueOf() {
        return this.primitiveValue ?? this.valueAsString;
    }
    equals(x) {
        if (!x)
            return false;
        return this.ref === x.ref;
    }
}
export class QuickJSScope extends QuickJSHandle {
    constructor(session, scopeInfo) {
        super(session, scopeInfo.reference);
        this.name = scopeInfo.name;
        this.type = 'object';
        this.primitive = false;
        this.isArray = false;
        this.expensive = scopeInfo.expensive;
    }
    toString() {
        return `[scope ${this.name}]`;
    }
}
export class QuickJSVariable extends QuickJSHandle {
    constructor(session, variableInfo) {
        super(session, variableInfo.variablesReference);
        this.name = variableInfo.name;
        this.type = variableInfo.type;
        this.primitive = true;
        this.isArray = false;
        switch (this.type) {
            case 'string':
                this.primitiveValue = variableInfo.value;
                break;
            case 'integer':
                this.primitiveValue = parseInt(variableInfo.value, 10);
                break;
            case 'float':
                this.primitiveValue = parseFloat(variableInfo.value);
                break;
            case 'boolean':
                this.primitiveValue = variableInfo.value === 'true';
                break;
            case 'null':
                this.primitiveValue = null;
                break;
            case 'undefined':
                this.primitiveValue = undefined;
                break;
            case 'object':
            case 'function':
                this.primitive = false;
                this.isArray = variableInfo.indexedVariables !== undefined;
                this.indexedCount = variableInfo.indexedVariables;
            // falls through
            default:
                this.valueAsString = variableInfo.value;
                this.primitive = false;
                break;
        }
    }
}
export class QuickJSDebugSession extends EventEmitter {
    constructor(connection) {
        super();
        this.connection = connection;
        connection.on('StoppedEvent', (ev) => {
            this.emit('stopped', ev);
        });
        connection.on('ThreadEvent', (ev) => {
            this.emit('context', ev);
        });
        connection.on('terminated', () => {
            this.emit('end');
        });
    }
    async continue() {
        return this.connection.sendRequest('continue');
    }
    async pause() {
        return this.connection.sendRequest('pause');
    }
    async stepNext() {
        return this.connection.sendRequest('next');
    }
    async stepIn() {
        return this.connection.sendRequest('stepIn');
    }
    async stepOut() {
        return this.connection.sendRequest('stepOut');
    }
    async evaluate(frameId, expression) {
        const res = await this.connection.sendRequest('evaluate', {
            frameId,
            expression
        });
        return new QuickJSVariable(this, { ...res, name: 'result', value: res.result });
    }
    async traceStack() {
        const res = await this.connection.sendRequest('stackTrace');
        return res.map((e) => new QuickJSStackFrame(this, e));
    }
    async getTopStack() {
        return (await this.traceStack())[0];
    }
    async getScopes(frameId) {
        const res = await this.connection.sendRequest('scopes', {
            frameId
        });
        return res.map((e) => new QuickJSScope(this, e));
    }
    async inspectVariable(reference, options) {
        const res = await this.connection.sendRequest('variables', {
            variablesReference: reference,
            ...options
        });
        return res.map((e) => new QuickJSVariable(this, e));
    }
    resume() {
        this.connection.sendMessage({
            type: 'resume',
            version: 1
        });
    }
    setBreakpoints(fileName, breakpoints) {
        this.connection.sendMessage({
            type: 'breakpoints',
            breakpoints: {
                path: fileName,
                breakpoints: breakpoints.length ? breakpoints : undefined
            },
            version: 1
        });
    }
    setStopOnException(enabled) {
        this.connection.sendMessage({
            type: 'stopOnException',
            stopOnException: enabled,
            version: 1
        });
    }
}
//# sourceMappingURL=session.js.map